from typing import Any, TypeVar
from collections.abc import Callable
from concurrent.futures import Future

T = TypeVar("T")

def f_apply(
    future_fn: Future[Callable[..., T]], *future_args: Future, **future_kwargs: Future
) -> Future[T]: ...



# -*- coding: utf-8 -*-

import math

from more_executors._impl.map import MapFuture
from more_executors._impl.common import MAX_TIMEOUT
from .check import ensure_future
from ..metrics import track_future


class ProxyFuture(MapFuture):
    # A future which proxies many calls through to the underlying object.
    def __init__(self, delegate, timeout):
        self.__timeout = timeout
        super(ProxyFuture, self).__init__(delegate)

    @property
    def __result(self):
        return self.result(self.__timeout)

    def __len__(self):
        return len(self.__result)

    def __getattr__(self, name):
        if name == "_ProxyFuture__result":
            # A rather wacky edge-case...
            # If self.__result is called, and the future has failed,
            # and the future's exception is an AttributeError:
            # python will interpret "AttributeError raised during self.__result"
            # as "I should try to use __getattr__ to get __result".
            # But we try to access self.__result again a few lines later
            # which would result in infinite recursion.
            # In that case, just raise the underlying exception (which we
            # know must be an AttributeError).
            raise self.exception()

        if name.startswith("__"):
            # Do not allow any special properties to trigger resolution
            # of the future, other than those we've explicitly proxied.
            raise AttributeError()

        return getattr(self.__result, name)

    def __getitem__(self, key):
        return self.__result[key]

    def __setitem__(self, key, value):
        self.__result[key] = value

    def __delitem__(self, key):
        del self.__result[key]

    def __iter__(self):
        return iter(self.__result)

    def __contains__(self, item):
        return item in self.__result

    def __add__(self, other):
        return self.__result + other

    def __sub__(self, other):
        return self.__result - other

    def __mul__(self, other):
        return self.__result * other

    def __div__(self, other):
        return self.__result.__div__(other)

    def __truediv__(self, other):
        return self.__result.__truediv__(other)

    def __floordiv__(self, other):
        return self.__result.__floordiv__(other)

    def __mod__(self, other):
        return self.__result % other

    def __divmod__(self, other):
        return divmod(self.__result, other)

    def __pow__(self, other, *modulo):
        return pow(self.__result, other, *modulo)

    def __lshift__(self, other):
        return self.__result << other

    def __rshift__(self, other):
        return self.__result >> other

    def __and__(self, other):
        return self.__result & other

    def __xor__(self, other):
        return self.__result ^ other

    def __or__(self, other):
        return self.__result | other

    def __neg__(self):
        return -self.__result  # pylint: disable=invalid-unary-operand-type

    def __pos__(self):
        return +self.__result  # pylint: disable=invalid-unary-operand-type

    def __abs__(self):
        return abs(self.__result)

    def __invert__(self):
        return ~self.__result  # pylint: disable=invalid-unary-operand-type

    def __complex__(self):
        return complex(self.__result)

    def __int__(self):
        return int(self.__result)

    def __float__(self):
        return float(self.__result)

    def __round__(self, *ndigits):
        return round(self.__result, *ndigits)

    def __trunc__(self):
        return self.__result.__trunc__()

    def __floor__(self):
        return math.floor(self.__result)

    def __ceil__(self):
        return math.ceil(self.__result)

    # Not proxied because I think it's more important for cases like this:
    #
    #  def some_fn(): Future|None
    #     ...
    #  f = some_fn()
    #  if f:
    #    # means some_fn returned a Future
    #  else:
    #    # means some_fn returned None
    #
    # ... to work as expected at first glance.
    #
    # Though not proxied, it has to be overridden (to return True for non-None)
    # because otherwise python checks if __len__() == 0.
    def __bool__(self):
        return True

    def __nonzero__(self):
        # python 2.x name for __bool__
        return self.__bool__()

    # Not proxied because nothing in Python itself uses it and I don't
    # want to spend the effort testing it
    # def __matmul__(self, other):

    # Not proxied because frankly I don't understand what this operator is
    # def __index__(self):

    # Not proxied because it seems unlikely anyone would implement a method
    # returning a future with a context manager.
    # def __enter__(self):
    # def __exit__(self, exc_type, exc_value, traceback):

    # String formatting functions and similar are not proxied as I'm
    # concerned about code like LOG.debug("Enqueued: %s", some_f) becoming
    # blocking only when the logger is enabled. Having major side-effects
    # on str() seems too risky.
    # def __str__(self):
    # def __bytes__(self):
    # def __format__(self, format_spec):

    # repr should never be proxied so that debugging code etc doesn't mislead
    # about the type of objects
    # def __repr__(self):

    # Not proxied so that:
    # - futures can be stored in dicts/lists which can be checked for membership
    #   without blocking
    # - no way to make it commutative?  e.g. we can make Future[123] == 123
    #   work but not 123 == Future[123].
    # def __eq__(self, other):

    # Not proxied so futures can be stored in dicts/sets without blocking
    # def __hash__(self):

    # Not proxied because eq is not proxied, which means expected relationship
    # between these operators can't be satisfied
    # def __ne__(self, other):
    # def __gt__(self, other):
    # def __ge__(self, other):
    # def __lt__(self, other):
    # def __le__(self, other):


@ensure_future
def f_proxy(f, **kwargs):
    """Proxy calls on a future through to the future's result.

    The value returned by this function is a future resolved with the
    same result (or exception) as the input future ``f``.  It will also
    proxy most attribute lookups and method calls through to the
    underlying result, awaiting the result when needed.

    Note that since the returned value is intended to remain usable as
    a ``Future``, this proxy is relatively conservative and avoids
    proxying functionality which would clash with the ``Future``
    interface.

    Functionality which is not proxied includes:

    - conversion to boolean (``__bool__``)
    - conversion to string (``__str__``, ``__repr__``)
    - methods relating to object identity (``__eq__``, ``__hash__``)

    Signature: :code:`Future<X> ⟶ Future<X>`

    Arguments:
        f (~concurrent.futures.Future)
            Any future.
        timeout (float)
            Timeout applied when awaiting the future's result
            during proxied calls.

    Returns:
        :class:`~concurrent.futures.Future`
            a Future which proxies calls through to the
            future's result as needed.

    .. versionadded:: 2.3.0
    """
    return track_future(
        ProxyFuture(f, timeout=kwargs.pop("timeout", MAX_TIMEOUT)), type="proxy"
    )



# -*- coding: utf-8 -*-

from threading import Lock
import logging
from concurrent.futures import Future

from .base import chain_cancel, weak_callback
from ..common import copy_future_exception, try_set_result
from .check import ensure_futures
from ..logwrap import LogWrapper
from ..metrics import track_future

LOG = LogWrapper(logging.getLogger("more_executors.futures"))


class BoolOperation(object):
    def __init__(self, fs):
        self.fs = {}
        for f in fs:
            self.fs[f] = True

        self.done = False
        self.lock = Lock()
        self.out = Future()

        for f in fs:
            chain_cancel(self.out, f)
            f.add_done_callback(weak_callback(self.handle_done))

    def get_state_update(self, f):
        raise NotImplementedError()  # pragma: no cover

    def handle_done(self, f):
        set_result = False
        set_exception = False
        cancel_futures = set()

        with self.lock:
            if self.done:
                return

            del self.fs[f]

            (set_result, set_exception, cancel_futures) = self.get_state_update(f)

        if set_result:
            try_set_result(self.out, f.result())
        if set_exception:
            copy_future_exception(f, self.out)

        for to_cancel in cancel_futures:
            to_cancel.cancel()


class OrOperation(BoolOperation):
    def get_state_update(self, f):
        set_result = False
        set_exception = False
        cancel_futures = set()

        # If it's the last result or it's a successful result:
        if (not self.fs) or (not f.cancelled() and not f.exception() and f.result()):
            self.done = True
            cancel_futures = list(self.fs.keys())
            if f.cancelled():
                # Cancelled => output is cancelled
                cancel_futures.append(self.out)
            elif f.exception():
                # Failed
                set_exception = True
            else:
                # Last or successful
                set_result = True

        return (set_result, set_exception, cancel_futures)


@ensure_futures
def f_or(f, *fs):
    """Boolean ``OR`` over a number of futures.

    Signature: :code:`Future<A>[, Future<B>[, ...]] ⟶ Future<A|B|...>`

    Arguments:
        f (~concurrent.futures.Future)
            Any future
        fs (~concurrent.futures.Future)
            Any futures

    Returns:
        :class:`~concurrent.futures.Future`
            A future resolved from the inputs using ``OR`` semantics:

            - Resolved with the earliest true value returned by an input
              future, if any.
            - Otherwise, resolved with the latest false value or exception
              returned by the input futures.

    .. note::
        This function is tested with up to 100,000 input futures.
        Exceeding this limit may result in performance issues.

    .. versionadded:: 1.19.0
    """
    if not fs:
        return f

    oper = OrOperation([f] + list(fs))
    track_future(oper.out, type="or")
    return oper.out


class AndOperation(BoolOperation):
    def get_state_update(self, f):
        set_result = False
        set_exception = False
        cancel_futures = set()

        if f.cancelled():
            # Cancelled => output is cancelled
            self.done = True
            cancel_futures = list(self.fs.keys())
            cancel_futures.append(self.out)
        elif f.exception():
            # Failed => we're done
            self.done = True
            set_exception = True
            cancel_futures = list(self.fs.keys())
        elif (not f.result()) or (not self.fs):
            # Falsey result or last result => we're done
            self.done = True
            cancel_futures = list(self.fs.keys())
            set_result = True

        return (set_result, set_exception, cancel_futures)


@ensure_futures
def f_and(f, *fs):
    """Boolean ``AND`` over a number of futures.

    Signature: :code:`Future<A>[, Future<B>[, ...]] ⟶ Future<A|B|...>`

    Arguments:
        f (~concurrent.futures.Future)
            Any future
        fs (~concurrent.futures.Future)
            Any futures

    Returns:
        :class:`~concurrent.futures.Future`
            A future resolved from the inputs using ``AND`` semantics:

            - Resolved with the latest value returned by an input
              future, if all futures are resolved with true values.
            - Otherwise, resolved with the earliest false value or exception
              returned by the input futures.

    .. note::
        This function is tested with up to 100,000 input futures.
        Exceeding this limit may result in performance issues.

    .. versionadded:: 1.19.0
    """
    if not fs:
        return f

    oper = AndOperation([f] + list(fs))
    track_future(oper.out, type="and")
    return oper.out



from concurrent.futures import Future
from typing import TypeVar
from collections.abc import Iterable, Callable

T = TypeVar("T")
A = TypeVar("A")
B = TypeVar("B")

def f_sequence(futures: Iterable[Future[T]]) -> Future[list[T]]: ...
def f_traverse(fn: Callable[[A], Future[B]], xs: Iterable[A]) -> Future[list[B]]: ...



from typing import Any, TypeVar
from concurrent.futures import Future

A = TypeVar("A")
B = TypeVar("B")

def f_or(f: Future[A], *fs: Future[B]) -> Future[A | B]: ...
def f_and(f: Future[A], *fs: Future[B]) -> Future[A | B]: ...



from .apply import f_apply as f_apply
from .base import (
    f_return as f_return,
    f_return_cancelled as f_return_cancelled,
    f_return_error as f_return_error,
)
from .bool import f_and as f_and, f_or as f_or
from .map import f_flat_map as f_flat_map, f_map as f_map
from .nocancel import f_nocancel as f_nocancel
from .proxy import f_proxy as f_proxy
from .sequence import f_sequence as f_sequence, f_traverse as f_traverse
from .timeout import f_timeout as f_timeout
from .zip import f_zip as f_zip



from typing import Any, Generic, TypeVar
from collections.abc import Callable

from .shared_types import ExecutorProtocol, TypedExecutorProtocol

A = TypeVar("A")
B = TypeVar("B")

class MapExecutor(TypedExecutorProtocol[A, B]):
    def __init__(
        self,
        delegate: ExecutorProtocol,
        fn: Callable[[A], B] = ...,
        logger: Any | None = ...,
        name: str = ...,
        **kwargs
    ) -> None: ...
    def __enter__(self) -> MapExecutor[A, B]: ...



from ._impl.futures import f_and as f_and
from ._impl.futures import f_apply as f_apply
from ._impl.futures import f_flat_map as f_flat_map
from ._impl.futures import f_map as f_map
from ._impl.futures import f_nocancel as f_nocancel
from ._impl.futures import f_or as f_or
from ._impl.futures import f_proxy as f_proxy
from ._impl.futures import f_return as f_return
from ._impl.futures import f_return_cancelled as f_return_cancelled
from ._impl.futures import f_return_error as f_return_error
from ._impl.futures import f_sequence as f_sequence
from ._impl.futures import f_timeout as f_timeout
from ._impl.futures import f_traverse as f_traverse
from ._impl.futures import f_zip as f_zip



from concurrent.futures import Future
from typing import Any, TypeVar
import logging
from collections.abc import Callable

from .shared_types import ExecutorProtocol, TypedExecutorProtocol

A = TypeVar("A")
B = TypeVar("B")
T = TypeVar("T")

class RetryPolicy(object): ...

class ExceptionRetryPolicy(RetryPolicy):
    def __init__(self, **kwargs): ...

class RetryExecutor(ExecutorProtocol):
    def __init__(
        self,
        delegate,
        retry_policy: Any | None = ...,
        logger: logging.Logger | None = ...,
        name: str = ...,
        **kwargs
    ): ...
    def submit_retry(
        self, retry_policy: RetryPolicy, fn: Callable[..., T], *args, **kwargs
    ) -> Future[T]: ...
    def __enter__(self) -> RetryExecutor: ...

class TypedRetryExecutor(TypedExecutorProtocol[A, B]):
    def submit_retry(
        self, retry_policy: RetryPolicy, fn: Callable[..., A], *args, **kwargs
    ) -> Future[B]: ...
    def __enter__(self) -> TypedRetryExecutor[A, B]: ...



from typing import TypeVar, Any
from concurrent.futures import Future

T = TypeVar("T")

def f_return(x: T) -> Future[T]: ...
def f_return_error(x: BaseException, traceback: Any = ...) -> Future[None]: ...
def f_return_cancelled() -> Future[None]: ...



# -*- coding: utf-8 -*-

from .base import wrap
from .check import ensure_future


@ensure_future
def f_map(future, fn=None, error_fn=None):
    """Map the output value of a future using the given functions.

    Signature: :code:`Future<A>, fn<A⟶B> ⟶ Future<B>`

    Arguments:
        future (~concurrent.futures.Future)
            Any future.
        fn (callable)
            Any callable to be applied on successful futures.
            This function is provided the result of the input future.
        error_fn (callable)
            Any callable to be applied on unsuccessful futures.
            This function is provided the exception of the input future.

    Returns:
        :class:`~concurrent.futures.Future`
            A future resolved with:

            - the returned value of :code:`fn(future.result())`
            - or the returned value of :code:`error_fn(future.exception())`
            - or with the exception raised by :obj:`future`, :obj:`fn` or :obj:`error_fn`.

    .. versionadded:: 1.19.0

    .. versionchanged:: 2.2.0
       Introduced ``error_fn``.
    """
    return wrap(future).with_map(fn=fn, error_fn=error_fn)()


@ensure_future
def f_flat_map(future, fn=None, error_fn=None):
    """Map the output value of a future using the given future-returning functions.

    Like :meth:`f_map`, except that the mapping functions must return a future,
    and that future will be flattened into the output value (avoiding a nested future).

    Signature: :code:`Future<A>, fn<A⟶Future<B>> ⟶ Future<B>`

    Arguments:
        future (~concurrent.futures.Future)
            Any future.
        fn (callable)
            Any future-returning callable to be applied on successful futures.
            This function is provided the result of the input future.
        error_fn (callable)
            Any future-returning callable to be applied on unsuccessful futures.
            This function is provided the exception of the input future.

    Returns:
        :class:`~concurrent.futures.Future`
            A future which is:

            - equivalent to that returned by :obj:`fn` if input future succeeded
            - or equivalent to that returned by :obj:`error_fn` if input future failed
            - or resolved with an exception if any of :obj:`future`, :obj:`fn` or
              :obj:`error_fn` failed

    .. versionadded:: 1.19.0

    .. versionchanged:: 2.2.0
       Introduced ``error_fn``.
    """
    return wrap(future).with_flat_map(fn=fn, error_fn=error_fn)()



import logging
from typing import TypeVar

from .shared_types import ExecutorProtocol, TypedExecutorProtocol

A = TypeVar("A")
B = TypeVar("B")

class CancelOnShutdownExecutor(ExecutorProtocol):
    def __init__(
        self,
        delegate: ExecutorProtocol,
        logger: logging.Logger | None = ...,
        name: str = ...,
    ) -> None: ...
    def __enter__(self) -> CancelOnShutdownExecutor: ...

class TypedCancelOnShutdownExecutor(TypedExecutorProtocol[A, B]):
    def __enter__(self) -> TypedCancelOnShutdownExecutor[A, B]: ...



"""A library of composable Python executors.

- [Documentation](https://rohanpm.github.io/more-executors/)
- [Source](https://github.com/rohanpm/more-executors)
- [PyPI](https://pypi.python.org/pypi/more-executors)

This library is intended for use with the
[`concurrent.futures`](https://docs.python.org/3/library/concurrent.futures.html)
module.  It includes a collection of `Executor` implementations in order to
extend the behavior of `Future` objects.

Compatible with Python 2.6, 2.7 and 3.x.

This documentation was built from an unknown revision.
"""
from ._impl.executors import Executors
from . import futures
from . import asyncio
from . import cancel_on_shutdown
from . import flat_map
from . import map  # pylint: disable=redefined-builtin
from . import poll
from . import retry
from . import sync
from . import throttle
from . import timeout

from .futures import *
from .asyncio import *
from .cancel_on_shutdown import *
from .flat_map import *
from .map import *
from .poll import *
from .retry import *
from .sync import *
from .throttle import *
from .timeout import *

__all__ = ["Executors", "futures"]
__all__.extend(futures.__all__)
__all__.extend(asyncio.__all__)
__all__.extend(cancel_on_shutdown.__all__)
__all__.extend(flat_map.__all__)
__all__.extend(map.__all__)
__all__.extend(poll.__all__)
__all__.extend(retry.__all__)
__all__.extend(sync.__all__)
__all__.extend(throttle.__all__)
__all__.extend(timeout.__all__)



"""Helper to get shutdown-aware instances of Event.

Call get_event to get instances of threading.Event.

These events work like any other event except that they will be automatically
set() during interpreter exit.  If an event has been set() for this reason,
is_shutdown() will return True.
"""

from threading import Event, RLock
import weakref
import atexit


class ShutdownAwareEventHandler(object):
    def __init__(self):
        self.lock = RLock()
        self.atexit_registered = False
        self.shutdown = False
        # NOTE: no weakset until python <2.7 is dropped
        self.events = []

    def clean_events(self, *_args, **_kwargs):
        with self.lock:
            self.events = [evt_ref for evt_ref in self.events if evt_ref()]

    def on_exiting(self):
        self.shutdown = True

        for evt_ref in self.events:
            evt = evt_ref()
            if evt:
                evt.set()

    def get_event(self):
        with self.lock:
            if not self.atexit_registered:
                atexit.register(self.on_exiting)
                self.atexit_registered = True
            out = Event()
            self.events.append(weakref.ref(out, self.clean_events))
            return out


GLOBAL_HANDLER = ShutdownAwareEventHandler()
get_event = GLOBAL_HANDLER.get_event


def is_shutdown():
    return GLOBAL_HANDLER.shutdown



from concurrent.futures import Executor, Future

from .common import copy_exception
from .wrap import CanCustomizeBind
from .metrics import metrics, track_future
from .helpers import ShutdownHelper


class SyncExecutor(CanCustomizeBind, Executor):
    """An executor which immediately invokes all submitted callables."""

    def __init__(self, logger=None, name="default"):
        """
        Parameters:
            logger (~logging.Logger):
                a logger used for messages from this executor
            name (str):
                a name for this executor

        .. versionchanged:: 2.7.0
            Introduced ``name``.
        """
        super(SyncExecutor, self).__init__()
        self._name = name
        self._shutdown = ShutdownHelper()
        metrics.EXEC_TOTAL.labels(type="sync", executor=self._name).inc()
        metrics.EXEC_INPROGRESS.labels(type="sync", executor=self._name).inc()

    def shutdown(self, wait=True, **_kwargs):
        if self._shutdown():
            super(SyncExecutor, self).shutdown(wait, **_kwargs)
            metrics.EXEC_INPROGRESS.labels(type="sync", executor=self._name).dec()

    def submit(self, fn, *args, **kwargs):  # pylint: disable=arguments-differ
        """Immediately invokes `fn(*args, **kwargs)` and returns a future
        with the result (or exception)."""
        with self._shutdown.ensure_alive():
            future = Future()
            track_future(future, type="sync", executor=self._name)

            try:
                result = fn(*args, **kwargs)
                future.set_result(result)
            except Exception:
                copy_exception(future)

            return future



from concurrent.futures import Executor
from threading import RLock
import logging

from .wrap import CanCustomizeBind
from .logwrap import LogWrapper
from .metrics import metrics
from .helpers import ShutdownHelper


class CancelOnShutdownExecutor(CanCustomizeBind, Executor):
    """An executor which delegates to another executor and cancels all
    futures when the executor is shut down.

    This class is useful in conjunction with executors having custom cancel
    behavior, such as :class:`~more_executors.PollExecutor`.

    .. note::
        From Python 3.9 onwards, the standard
        :meth:`~concurrent.futures.Executor.shutdown` method includes a
        ``cancel_futures`` parameter which can be used to cancel futures
        on shutdown.

        These two approaches of cancelling futures on shutdown have the
        following differences:

        - ``cancel_futures=True`` will only cancel futures which have not
          yet started running.
        - ``CancelOnShutdownExecutor`` will attempt to cancel *all* incomplete
          futures, including those which are running.

        If you're using other executors from this library and you want to
        ensure futures are cancelled on shutdown, ``CancelOnShutdownExecutor``
        should be preferred because several of the executor classes in this
        library do support cancellation of running futures (unlike the executors
        in the standard library).
    """

    def __init__(self, delegate, logger=None, name="default"):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                executor to which callables will be submitted

            logger (~logging.Logger):
                a logger used for messages from this executor

            name (str):
                a name for this executor

        .. versionchanged:: 2.7.0
            Introduced ``name``.
        """
        self._log = LogWrapper(
            logger if logger else logging.getLogger("CancelOnShutdownExecutor")
        )
        self._name = name
        self._delegate = delegate
        self._futures = set()
        self._lock = RLock()
        self._shutdown = ShutdownHelper()
        metrics.EXEC_TOTAL.labels(type="cancel_on_shutdown", executor=self._name).inc()
        metrics.EXEC_INPROGRESS.labels(
            type="cancel_on_shutdown", executor=self._name
        ).inc()

    def shutdown(self, wait=True, **_kwargs):
        """Shut down the executor.

        All futures created by this executor which have not yet been completed
        will have :meth:`~concurrent.futures.Future.cancel` invoked.

        Note that there is no guarantee that the cancel will succeed, and only a single
        attempt is made to cancel any future.
        """
        with self._lock:
            if not self._shutdown():
                return
            metrics.EXEC_INPROGRESS.labels(
                type="cancel_on_shutdown", executor=self._name
            ).dec()
            futures = self._futures.copy()

        for f in futures:
            cancel = f.cancel()
            self._log.debug("Cancel %s: %s", f, cancel)
            if cancel:
                metrics.SHUTDOWN_CANCEL.labels(executor=self._name).inc()

        self._delegate.shutdown(wait, **_kwargs)

    def submit(self, *args, **kwargs):  # pylint: disable=arguments-differ
        with self._shutdown.ensure_alive():
            with self._lock:
                future = self._delegate.submit(*args, **kwargs)
                self._futures.add(future)
                future.add_done_callback(self._futures.discard)
            return future



from concurrent.futures import Executor
from threading import RLock, Thread
import weakref
import logging

try:
    from time import monotonic
except ImportError:  # pragma: no cover
    from monotonic import monotonic

from .common import (
    _Future,
    MAX_TIMEOUT,
    copy_exception,
    copy_future_exception,
    try_set_result,
)
from .wrap import CanCustomizeBind
from .helpers import executor_loop
from .event import get_event, is_shutdown
from .logwrap import LogWrapper
from .helpers import ShutdownHelper
from .metrics import metrics, track_future


class PollFuture(_Future):
    def __init__(self, delegate, executor):
        super(PollFuture, self).__init__()
        self._delegate = delegate
        self._executor = executor
        self._delegate.add_done_callback(self._delegate_resolved)
        self.add_done_callback(self._clear_executor)

    def _delegate_resolved(self, delegate):
        assert delegate is self._delegate, "BUG: called with %s, expected %s" % (
            delegate,
            self._delegate,
        )

        if delegate.cancelled():
            return
        if delegate.exception():
            copy_future_exception(delegate, self)
        else:
            self._executor._register_poll(self, self._delegate)

    def _clear_delegate(self):
        with self._me_lock:
            self._delegate = None

    @classmethod
    def _clear_executor(cls, future):
        future._executor._deregister_poll(future)
        future._executor = None

    def set_result(self, result):
        with self._me_lock:
            if self.done():
                return
            super(PollFuture, self).set_result(result)
        self._me_invoke_callbacks()

    def set_exception(self, exception):
        with self._me_lock:
            # We can't get here because we always try set_exception_info first,
            # which succeeds if done() is true, in which case we won't fall
            # back to set_exception.
            # if self.done():
            #    return
            super(PollFuture, self).set_exception(exception)
        self._me_invoke_callbacks()

    def set_exception_info(self, exception, traceback):
        # For python2 compat.
        # pylint: disable=no-member
        with self._me_lock:
            if self.done():
                return
            super(PollFuture, self).set_exception_info(exception, traceback)
        self._me_invoke_callbacks()

    def running(self):
        with self._me_lock:
            if self._delegate:
                return self._delegate.running()
        # If delegate is removed, we're now polling or done.
        return False

    def _me_cancel(self):
        if self._delegate and not self._delegate.cancel():
            return False
        executor = self._executor
        return executor and executor._run_cancel_fn(self)


class PollDescriptor(object):
    """Represents an unresolved :class:`~concurrent.futures.Future`.

    The poll function used by :class:`PollExecutor` will be
    invoked with a list of PollDescriptor objects.
    """

    def __init__(self, future, result):
        self.__future = future
        self.__result = result

    @property
    def result(self):
        """The result from the delegate executor's future, which should be used to
        drive the poll."""
        return self.__result

    def yield_result(self, result):
        """The poll function can call this function to make the future yield the given result.

        Arguments:
            result (object):
                a result to be returned by the future associated with this descriptor
        """
        try_set_result(self.__future, result)

    def yield_exception(self, exception, traceback=None):
        """The poll function can call this function to make the future raise the given exception.

        Arguments:
           exception (Exception):
               An exception to be returned or raised by the future associated with this
               descriptor
           traceback (traceback):
               An optional associated traceback. This argument only has an effect on python 2.x,
               where exception objects do not include a traceback.
        """

        copy_exception(self.__future, exception, traceback)


class PollExecutor(CanCustomizeBind, Executor):
    """Instances of `PollExecutor` submit callables to a delegate executor
    and resolve the returned futures via a provided poll function.

    A cancel function may also be provided to perform additional processing
    when a returned future is cancelled.
    """

    def __init__(
        self,
        delegate,
        poll_fn,
        cancel_fn=None,
        default_interval=5.0,
        logger=None,
        name="default",
    ):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                executor to which callables will be submitted

            poll_fn (callable):
                a `poll function`_ used to decide when futures should be resolved

            cancel_fn (callable):
                a `cancel function`_ invoked when future cancel is required

            default_interval (float):
                default interval between polls (in seconds)

            logger (~logging.Logger):
                a logger used for messages from this executor

            name (str):
                a name for this executor

        .. versionchanged:: 2.7.0
            Introduced ``name``.
        """
        self._log = LogWrapper(logger if logger else logging.getLogger("PollExecutor"))
        self._name = name
        self._delegate = delegate
        self._default_interval = default_interval
        self._poll_fn = poll_fn
        self._cancel_fn = cancel_fn
        self._poll_descriptors = []
        self._poll_event = get_event()

        poll_event = self._poll_event
        self_ref = weakref.ref(self, lambda _: poll_event.set())

        self._poll_thread = Thread(
            name="PollExecutor-%s" % name, target=_poll_loop, args=(self_ref,)
        )
        self._poll_thread.daemon = True
        self._shutdown = ShutdownHelper()
        self._lock = RLock()

        metrics.EXEC_TOTAL.labels(type="poll", executor=self._name).inc()
        metrics.EXEC_INPROGRESS.labels(type="poll", executor=self._name).inc()

        self._poll_thread.start()

    def submit(self, *args, **kwargs):  # pylint: disable=arguments-differ
        with self._shutdown.ensure_alive():
            delegate_future = self._delegate.submit(*args, **kwargs)
            out = PollFuture(delegate_future, self)
            track_future(out, type="poll", executor=self._name)
            return out

    def notify(self):
        """Request the executor to re-run the polling function as soon as possible.

        This method may be used to perform the next poll earlier than it would
        otherwise run.

        This is useful for cases where futures may be resolved via a mixture of
        polling and external events.

        .. versionadded:: 2.2.0
        """
        self._poll_event.set()

    def _register_poll(self, future, delegate_future):
        descriptor = PollDescriptor(future, delegate_future.result())
        with self._lock:
            self._poll_descriptors.append((future, descriptor))
            future._clear_delegate()
            self._poll_event.set()

    def _deregister_poll(self, future):
        with self._lock:
            self._poll_descriptors = [
                (f, d) for (f, d) in self._poll_descriptors if f is not future
            ]

    def _run_cancel_fn(self, future):
        if not self._cancel_fn:
            # no cancel function => no veto of cancel
            return True

        descriptor = [d for (f, d) in self._poll_descriptors if f is future]
        if not descriptor:
            # no record of this future => no veto of cancel.
            # we can get here if the future is already done
            # or if polling hasn't started yet
            return True

        assert len(descriptor) == 1, "Too many poll descriptors for %s" % future

        descriptor = descriptor[0]

        try:
            return self._cancel_fn(descriptor.result)
        except Exception:
            self._log.exception(
                "Exception during cancel on %s/%s", future, descriptor.result
            )
            return False

    def _run_poll_fn(self):
        with self._lock:
            descriptors = [d for (_, d) in self._poll_descriptors]

        try:
            now = monotonic()
            return self._poll_fn(descriptors)
        except Exception as e:
            self._log.debug("Poll function failed", exc_info=True)
            metrics.POLL_ERROR.labels(executor=self._name).inc()
            # If poll function fails, then every future
            # depending on the poll also immediately fails.
            [d.yield_exception(e) for d in descriptors]
        finally:
            metrics.POLL_TOTAL.labels(executor=self._name).inc()
            metrics.POLL_TIME.labels(executor=self._name).inc(monotonic() - now)

    def shutdown(self, wait=True, **_kwargs):
        if self._shutdown():
            metrics.EXEC_INPROGRESS.labels(type="poll", executor=self._name).dec()
            self._poll_event.set()
            self._delegate.shutdown(wait, **_kwargs)
            if wait:
                self._log.debug("Join poll thread...")
                self._poll_thread.join(MAX_TIMEOUT)
                self._log.debug("Joined poll thread.")


@executor_loop
def _poll_loop(executor_ref):
    while True:
        executor = executor_ref()
        if not executor:
            break

        if executor._shutdown.is_shutdown or is_shutdown():
            break

        executor._log.debug("Polling...")

        next_sleep = executor._run_poll_fn()
        if not (isinstance(next_sleep, int) or isinstance(next_sleep, float)):
            next_sleep = executor._default_interval

        executor._log.debug("Sleeping...")

        poll_event = executor._poll_event
        del executor

        poll_event.wait(next_sleep)
        poll_event.clear()



import os
import logging
from functools import partial

try:
    from time import monotonic
except ImportError:  # pragma: no cover
    from monotonic import monotonic

from .null import NullMetrics

LOG = logging.getLogger("more-executors.metrics")

if os.environ.get("MORE_EXECUTORS_PROMETHEUS", "1") == "0":
    metrics = NullMetrics()  # type: ignore
else:
    try:  # pylint: disable=import-error
        from .prometheus import PrometheusMetrics

        metrics = PrometheusMetrics()
    except Exception:
        LOG.debug("disabling prometheus support", exc_info=True)

        metrics = NullMetrics()  # type: ignore


def record_done(f, started_when, time, inprogress, cancelled, failed):
    inprogress.dec()

    run_time = monotonic() - started_when
    time.inc(run_time)

    if f.cancelled():
        cancelled.inc()
    elif f.exception():
        failed.inc()


def track_future(f, **labels):
    if "executor" not in labels:
        # TODO: when using the future composition functions like
        # f_or, f_map etc, we'll often get here. But it might be possible
        # to do something better? e.g. if we have one future chained from
        # another, can we copy the label from the other?
        labels["executor"] = "default"

    metrics.FUTURE_TOTAL.labels(**labels).inc()

    start = monotonic()

    inprogress = metrics.FUTURE_INPROGRESS.labels(**labels)
    inprogress.inc()

    time = metrics.FUTURE_TIME.labels(**labels)

    cancelled = metrics.FUTURE_CANCEL.labels(**labels)
    failed = metrics.FUTURE_ERROR.labels(**labels)
    cb = partial(
        record_done,
        started_when=start,
        time=time,
        inprogress=inprogress,
        cancelled=cancelled,
        failed=failed,
    )

    f.add_done_callback(cb)

    return f


def track_future_noop(f, **_labels):
    return f


if isinstance(metrics, NullMetrics):
    track_future = track_future_noop



class NullBase(object):
    def labels(self, **_kwargs):
        return self

    def inc(self, _value=1):
        pass


class Counter(NullBase):
    pass


class Gauge(NullBase):
    def dec(self, _value=1):
        pass


class NullMetrics(object):
    TIMEOUT = Counter()
    SHUTDOWN_CANCEL = Counter()
    EXEC_INPROGRESS = Gauge()
    EXEC_TOTAL = Counter()
    FUTURE_INPROGRESS = Gauge()
    FUTURE_TOTAL = Counter()
    FUTURE_CANCEL = Counter()
    FUTURE_ERROR = Counter()
    FUTURE_TIME = Counter()
    POLL_TOTAL = Counter()
    POLL_ERROR = Counter()
    POLL_TIME = Counter()
    RETRY_TOTAL = Counter()
    RETRY_QUEUE = Gauge()
    RETRY_DELAY = Counter()
    THROTTLE_QUEUE = Gauge()



# -*- coding: utf-8 -*-
from concurrent.futures import Future
from threading import Lock
from functools import partial
from collections import namedtuple

from more_executors._impl.common import (
    copy_future_exception,
    try_set_result,
)
from .base import f_return, chain_cancel, weak_callback
from .check import ensure_futures
from ..metrics import track_future


# For small-ish tuples, we make f_zip return namedtuple instances
# which can be traced back to here rather than bare tuples. The point
# is to improve debuggability of code where the future returned by
# f_zip ends up being logged with repr(), as in that case only the
# result's type will be logged. Seeing e.g. 'ZipTuple3' in a crash log
# rather than 'tuple' could be potentially very helpful.
TUPLE_CLASSES = []
for i in range(0, 20):
    TUPLE_CLASSES.append(
        namedtuple("ZipTuple%s" % i, ["f%s" % idx for idx in range(0, i)])
    )


def maketuple(value):
    vlen = len(value)
    if vlen < len(TUPLE_CLASSES):
        return TUPLE_CLASSES[vlen](*value)
    return tuple(value)


class Zipper(object):
    def __init__(self, fs):
        self.fs = list(fs)
        self.out = Future()
        self.done = False
        self.lock = Lock()
        self.count_remaining = len(self.fs)

        for (idx, future) in enumerate(self.fs):
            chain_cancel(self.out, future)
            future.add_done_callback(weak_callback(partial(self.handle_done, idx)))

    def handle_done(self, index, f):
        set_result = False
        set_exception = False
        cancel = False

        with self.lock:
            if self.done:
                pass
            elif f.cancelled():
                self.done = True
                cancel = True
            elif f.exception():
                self.done = True
                set_exception = True
            else:
                self.fs[index] = f.result()
                self.count_remaining -= 1

                if self.count_remaining == 0:
                    self.done = True
                    set_result = True

        if cancel:
            self.out.cancel()
        if set_result:
            try_set_result(self.out, maketuple(self.fs))
        if set_exception:
            copy_future_exception(f, self.out)


@ensure_futures
def f_zip(*fs):
    """Create a new future holding the return values of any number of input futures.

    Signature: :code:`Future<A>[, Future<B>[, ...]] ⟶ Future<A[, B[, ...]]>`

    Arguments:
        fs (~concurrent.futures.Future)
            Any number of futures.

    Returns:
        :class:`~concurrent.futures.Future` of :class:`tuple`
            A future holding the returned values of all input futures as a tuple.
            The returned tuple has the same length and order as the input futures.

            Alternatively, a future raising an exception or a cancelled future,
            if any input futures raised an exception or was cancelled.

    .. note::
        This function is tested with up to 100,000 input futures.
        Exceeding this limit may result in performance issues.

    .. versionadded:: 1.19.0
    """
    if not fs:
        return f_return(maketuple([]))

    return track_future(Zipper(fs).out, type="zip")



import logging
from functools import wraps
from threading import Lock
from contextlib import contextmanager

from .logwrap import LogWrapper

LOG = LogWrapper(logging.getLogger(__name__))


def executor_loop(fn):
    @wraps(fn)
    def out(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except RuntimeError as error:
            if "cannot schedule new futures after" in str(error):
                LOG.debug("Ignoring error due to interpreter shutdown", exc_info=1)
                return
            raise

    return out


class ShutdownHelper(object):
    def __init__(self):
        self._lock = Lock()
        self.is_shutdown = False

    @contextmanager
    def ensure_alive(self):
        with self._lock:
            if self.is_shutdown:
                raise RuntimeError("cannot schedule new futures after shutdown")
            yield

    def __call__(self):
        # Ensure shut down, return True if newly shutdown or
        # False if already shutdown
        with self._lock:
            if self.is_shutdown:
                return False
            self.is_shutdown = True
            return True



from typing import TypeVar, Any
import logging
from collections.abc import Callable
from concurrent.futures import Future

from .shared_types import ExecutorProtocol, TypedExecutorProtocol

A = TypeVar("A")
B = TypeVar("B")

class ThrottleExecutor(ExecutorProtocol):
    def __init__(
        self,
        delegate: ExecutorProtocol,
        count: int | Callable[[], int],
        logger: logging.Logger | None = ...,
        name: str = ...,
        block: bool = ...,
    ): ...
    def __enter__(self) -> ThrottleExecutor: ...

class TypedThrottleExecutor(TypedExecutorProtocol[A, B]):
    def __init__(
        self,
        delegate: ExecutorProtocol,
        count: int | Callable[[], int],
        logger: logging.Logger | None = ...,
        name: str = ...,
        block: bool = ...,
    ): ...
    def __enter__(self) -> TypedThrottleExecutor[A, B]: ...



import os


class LogWrapper(object):
    # A helper to veto all debug logs unless MORE_EXECUTORS_DEBUG is set.
    #
    # The point here is that, even for DEBUG level, we can be a bit too verbose.
    def __init__(self, logger):
        self._delegate = logger
        self._debug = os.environ.get("MORE_EXECUTORS_DEBUG", "0") == "1"
        self.info = logger.info
        self.exception = logger.exception

    def debug(self, *args, **kwargs):
        if self._debug:
            return self._delegate.debug(*args, **kwargs)



from typing import TypeVar, Any
import logging
from collections.abc import Callable
from concurrent.futures import Future

from .shared_types import ExecutorProtocol, TypedExecutorProtocol

T = TypeVar("T")
A = TypeVar("A")
B = TypeVar("B")

class TimeoutExecutor(ExecutorProtocol):
    def __init__(
        self,
        delegate: ExecutorProtocol | TypedExecutorProtocol[object, object],
        timeout: float | int,
        logger: logging.Logger | None = ...,
        name: str = ...,
    ): ...
    def submit_timeout(
        self, timeout: float | int, fn: Callable[..., T], *args, **kwargs
    ) -> Future[T]: ...
    def __enter__(self) -> TimeoutExecutor: ...

class TypedTimeoutExecutor(TypedExecutorProtocol[A, B]):
    def submit_timeout(
        self, timeout: float | int, fn: Callable[..., A], *args, **kwargs
    ) -> Future[B]: ...
    def __enter__(self) -> TypedTimeoutExecutor[A, B]: ...



from typing import Protocol, TypeVar, Any
from collections.abc import Callable
from concurrent.futures import Future

# NOTE: we have circular dependencies here,
# many of the return types from ExecutorProtocol methods themselves inherit
# ExecutorProtocol and hence import this file. Nevertheless it seems to work,
# and using string annotations sometimes makes things *not* work (mostly
# tested in vscode/pylance). Not sure why, maybe .pyi works differently from
# other imports - can't find info about it.

from .cancel_on_shutdown import CancelOnShutdownExecutor, TypedCancelOnShutdownExecutor
from .retry import RetryExecutor, TypedRetryExecutor
from .map import MapExecutor
from .flat_map import FlatMapExecutor
from .timeout import TimeoutExecutor, TypedTimeoutExecutor
from .throttle import TypedThrottleExecutor, ThrottleExecutor
from .poll import PollDescriptor, PollExecutor
from .asyncio import AsyncioExecutor

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
T = TypeVar("T")

class ExecutorProtocol(Protocol):
    # Protocol for typical executor where submit output depends on the
    # return type of callables.
    def submit(self, fn: Callable[..., T], *args, **kwargs) -> Future[T]: ...
    def bind(self, *args, **kwargs): ...
    def flat_bind(self, *args, **kwargs): ...
    def with_retry(self, *args, **kwargs) -> RetryExecutor: ...
    def with_map(self, fn: Callable[[A], B], *args, **kwargs) -> MapExecutor[A, B]: ...
    def with_flat_map(
        self, fn: Callable[[A], Future[B]], *args, **kwargs
    ) -> FlatMapExecutor[A, B]: ...
    def with_poll(
        self, poll_fn: Callable[[list[PollDescriptor[A, B]]], Any], *args, **kwargs
    ) -> PollExecutor[A, B]: ...
    def with_timeout(self, *args, **kwargs) -> TimeoutExecutor: ...
    def with_throttle(self, *args, **kwargs) -> ThrottleExecutor: ...
    def with_cancel_on_shutdown(self, *args, **kwargs) -> CancelOnShutdownExecutor: ...
    def with_asyncio(self, *args, **kwargs) -> AsyncioExecutor: ...
    def __exit__(self, *args, **kwargs) -> None: ...
    def __enter__(self) -> ExecutorProtocol: ...

class TypedExecutorProtocol(Protocol[A, B]):
    # Protocol for executor where callable and output types are known.
    def submit(self, fn: Callable[..., A], *args, **kwargs) -> Future[B]: ...
    def with_retry(self, *args, **kwargs) -> TypedRetryExecutor[A, B]: ...
    def with_map(self, fn: Callable[[B], C], *args, **kwargs) -> MapExecutor[A, C]: ...
    def with_flat_map(
        self, fn: Callable[[B], Future[C]], *args, **kwargs
    ) -> FlatMapExecutor[A, C]: ...
    def with_poll(
        self, poll_fn: Callable[[list[PollDescriptor[B, C]]], Any], *args, **kwargs
    ) -> PollExecutor[A, C]: ...
    def with_timeout(self, *args, **kwargs) -> TypedTimeoutExecutor[A, B]: ...
    def with_throttle(self, *args, **kwargs) -> TypedThrottleExecutor[A, B]: ...
    def with_cancel_on_shutdown(
        self, *args, **kwargs
    ) -> TypedCancelOnShutdownExecutor[A, B]: ...
    def with_asyncio(self, *args, **kwargs) -> AsyncioExecutor: ...
    def __exit__(self, *args, **kwargs) -> None: ...
    def __enter__(self) -> TypedExecutorProtocol[A, B]: ...



from concurrent.futures import Executor
from threading import Thread, Lock
from collections import namedtuple, deque
from functools import partial
import logging
import weakref

from .common import MAX_TIMEOUT
from .wrap import CanCustomizeBind
from .map import MapFuture
from .helpers import executor_loop, ShutdownHelper
from .event import get_event, is_shutdown
from .logwrap import LogWrapper

from .metrics import metrics, track_future


class ThrottleFuture(MapFuture):
    def __init__(self, executor):
        self._executor = executor
        super(ThrottleFuture, self).__init__(delegate=None, map_fn=lambda x: x)
        self.add_done_callback(self._clear_executor)

    def _me_cancel(self):
        if self._delegate:
            return self._delegate.cancel()
        executor = self._executor
        return executor and executor._do_cancel(self)

    @classmethod
    def _clear_executor(cls, future):
        future._executor = None


ThrottleJob = namedtuple("ThrottleJob", ["future", "fn", "args", "kwargs"])


class AtomicInt(object):
    def __init__(self):
        self.value = 0
        self.lock = Lock()

    def decr(self):
        with self.lock:
            self.value -= 1

    def incr(self):
        with self.lock:
            self.value += 1


class ThrottleExecutor(CanCustomizeBind, Executor):
    """An executor which delegates to another executor while enforcing
    a limit on the number of futures running concurrently.

    - Callables are submitted to the delegate executor, from a different
      thread than the calling thread.

    - Where `count` is used to initialize this executor, if there
      are already `count` futures submitted to the delegate executor and not
      yet :meth:`~concurrent.futures.Future.done`, additional callables will
      either be queued or will block on submit, and will only be submitted
      to the delegate executor once there are less than `count` futures in
      progress.

    .. versionadded:: 1.9.0
    """

    def __init__(self, delegate, count, logger=None, name="default", block=False):
        """
        Parameters:
            delegate (~concurrent.futures.Executor):
                an executor to which callables will be submitted

            count (int, callable):
                int:
                    maximum number of concurrently running futures.
                callable:
                    a callable which returns an ``int`` (or ``None``, to indicate
                    no throttling).

                    The callable will be invoked each time this executor needs
                    to decide whether to throttle futures; this may be used
                    to implement dynamic throttling.

                    .. versionadded:: 2.5.0

            block (bool)
                If ``True``, calls to ``submit()`` on this executor may block if
                there are already ``count`` futures in progress.

                Otherwise, calls to ``submit()`` will always return immediately
                and callables will be queued internally.

            logger (~logging.Logger):
                a logger used for messages from this executor

            name (str):
                a name for this executor

        .. versionchanged:: 2.7.0
            Introduced ``name``.

        .. versionchanged:: 2.11.0
            Introduced ``block``.
        """
        self._log = LogWrapper(
            logger if logger else logging.getLogger("ThrottleExecutor")
        )
        self._block = block
        self._name = name
        self._delegate = delegate
        self._to_submit = deque()
        self._lock = Lock()
        self._event = get_event()
        self._running_count = AtomicInt()
        self._throttle = count if callable(count) else lambda: count
        self._last_throttle = self._throttle()
        self._shutdown = ShutdownHelper()

        event = self._event
        self_ref = weakref.ref(self, lambda _: event.set())

        metrics.EXEC_INPROGRESS.labels(type="throttle", executor=self._name).inc()
        metrics.EXEC_TOTAL.labels(type="throttle", executor=self._name).inc()

        self._thread = Thread(
            name="ThrottleExecutor-%s" % name, target=_submit_loop, args=(self_ref,)
        )
        self._thread.daemon = True
        self._thread.start()

    def submit(self, fn, *args, **kwargs):  # pylint: disable=arguments-differ
        with self._shutdown.ensure_alive():
            self._block_until_ready(self._eval_throttle())

            out = ThrottleFuture(self)
            track_future(out, type="throttle", executor=self._name)

            job = ThrottleJob(out, fn, args, kwargs)
            with self._lock:
                self._to_submit.append(job)
                metrics.THROTTLE_QUEUE.labels(executor=self._name).inc()
                self._log.debug("Enqueued: %s", job)
            self._event.set()
            return out

    def shutdown(self, wait=True, **_kwargs):
        if self._shutdown():
            self._log.debug("Shutting down")
            metrics.EXEC_INPROGRESS.labels(type="throttle", executor=self._name).dec()
            self._delegate.shutdown(wait, **_kwargs)
            self._event.set()
            if wait:
                self._thread.join(MAX_TIMEOUT)

    def _block_until_ready(self, throttle_val):
        while self._block and not self._shutdown.is_shutdown:
            if len(self._to_submit) < throttle_val:
                return
            self._log.debug("%s: throttling on submit", self._name)
            self._event.wait(30.0)

    def _eval_throttle(self):
        try:
            self._last_throttle = self._throttle()
        except Exception:
            self._log.exception(
                "Error evaluating throttle count via %r", self._throttle
            )

        return self._last_throttle

    def _do_submit(self, job):
        delegate_future = self._delegate.submit(job.fn, *job.args, **job.kwargs)
        self._log.debug("Submitted %s yielding %s", job, delegate_future)

        delegate_future.add_done_callback(
            partial(
                self._delegate_future_done, self._log, self._running_count, self._event
            )
        )
        job.future._set_delegate(delegate_future)

    def _do_cancel(self, future):
        with self._lock:
            for job in self._to_submit:
                if job.future is future:
                    self._to_submit.remove(job)
                    self._log.debug("Cancelled %s", job)
                    return True
        self._log.debug("Could not find for cancel: %s", future)
        return False

    @classmethod
    def _delegate_future_done(cls, log, running_count, event, future):
        # Whenever an inner future completes, the thread should wake up
        # in case there's something to be submitted
        log.debug("Delegate future done: %s", future)
        running_count.decr()
        event.set()


def _submit_loop_iter(executor):
    if not executor:
        return

    if executor._shutdown.is_shutdown or is_shutdown():
        return

    throttle = executor._eval_throttle()
    to_submit = []
    with executor._lock:
        while executor._to_submit:
            if throttle is not None and (executor._running_count.value >= throttle):
                executor._log.debug("Throttled")
                break
            job = executor._to_submit.popleft()
            executor._log.debug("Will submit: %s", job)
            to_submit.append(job)

            # While not actually running yet, we've committed to running it, so...
            executor._running_count.incr()
            metrics.THROTTLE_QUEUE.labels(executor=executor._name).dec()

        executor._log.debug(
            "Submitting %s, throttling %s", len(to_submit), len(executor._to_submit)
        )

    for job in to_submit:
        executor._do_submit(job)

    # Because the throttle count is dynamic, we should wake up at some point in
    # the future to re-check throttle, even if no other events occurred.
    # If there's nothing running at all, we should do this sooner, because
    # there won't be any events from completing futures.
    return executor._event, 30.0 if executor._running_count.value else 2.0


@executor_loop
def _submit_loop(executor_ref):
    while True:
        result = _submit_loop_iter(executor_ref())
        if not result:
            break

        (event, wait_time) = result
        event.wait(wait_time)
        event.clear()



from ._impl.executors import Executors as Executors

from .futures import *
from .asyncio import *
from .cancel_on_shutdown import *
from .flat_map import *
from .map import *
from .poll import *
from .retry import *
from .sync import *
from .throttle import *
from .timeout import *



from functools import partial

import prometheus_client  # pylint: disable=import-error


Counter = partial(prometheus_client.Counter, namespace="more_executors")
Gauge = partial(prometheus_client.Gauge, namespace="more_executors")


class PrometheusMetrics(object):
    TIMEOUT = Counter(
        "timeout", "Futures cancelled due to timeout", labelnames=("executor",)
    )
    SHUTDOWN_CANCEL = Counter(
        "shutdown_cancel", "Futures cancelled at shutdown", labelnames=("executor",)
    )
    EXEC_INPROGRESS = Gauge(
        "exec_inprogress", "Executors currently in use", labelnames=("type", "executor")
    )
    EXEC_TOTAL = Counter(
        "exec_total", "Total executors used", labelnames=("type", "executor")
    )
    FUTURE_INPROGRESS = Gauge(
        "future_inprogress", "Futures currently in use", labelnames=("type", "executor")
    )
    FUTURE_TOTAL = Counter(
        "future_total", "Total futures used", labelnames=("type", "executor")
    )
    FUTURE_CANCEL = Counter(
        "future_cancel", "Futures cancelled", labelnames=("type", "executor")
    )
    FUTURE_ERROR = Counter(
        "future_error", "Futures resolved with error", labelnames=("type", "executor")
    )
    FUTURE_TIME = Counter(
        "future_time",
        "Total execution time of futures",
        labelnames=("type", "executor"),
    )
    POLL_TOTAL = Counter(
        "poll_total", "PollExecutor's poll function calls", labelnames=("executor",)
    )
    POLL_ERROR = Counter(
        "poll_error",
        "PollExecutor's poll function calls raising an error",
        labelnames=("executor",),
    )
    POLL_TIME = Counter(
        "poll_time",
        "Total execution time of PollExecutor's poll function",
        labelnames=("executor",),
    )
    RETRY_TOTAL = Counter(
        "retry_total", "Futures retried by RetryExecutor", labelnames=("executor",)
    )
    RETRY_QUEUE = Gauge(
        "retry_queue", "Futures in RetryExecutor queue", labelnames=("executor",)
    )
    RETRY_DELAY = Counter(
        "retry_delay", "Time spent waiting for retry", labelnames=("executor",)
    )
    # TODO: figure out how to implement this one
    # THROTTLE_TOTAL = Counter("throttle_total", "Futures throttled by ThrottleExecutor")
    THROTTLE_QUEUE = Gauge(
        "throttle_queue", "Futures in ThrottleExecutor queue", labelnames=("executor",)
    )



from concurrent.futures import Executor
from typing import Any, Generic, TypeVar
from collections.abc import Callable
import logging

from .shared_types import ExecutorProtocol, TypedExecutorProtocol

A = TypeVar("A")
B = TypeVar("B")

class PollDescriptor(Generic[A, B]):
    @property
    def result(self) -> A: ...
    def yield_result(self, result: B) -> None: ...
    def yield_exception(
        self, exception: BaseException, traceback: Any | None = ...
    ) -> None: ...

class PollExecutor(TypedExecutorProtocol[A, B]):
    def __init__(
        self,
        delegate: ExecutorProtocol,
        poll_fn: Callable[[list[PollDescriptor[A, B]]], int | float | None],
        cancel_fn: Callable[[A], bool] | None = ...,
        default_interval: float = ...,
        logger: logging.Logger | None = ...,
        name: str = ...,
    ): ...
    def notify(self) -> None: ...
    def __enter__(self) -> PollExecutor[A, B]: ...



from .metrics import metrics as metrics
from concurrent.futures import Executor
from typing import Any

class AsyncioExecutor(Executor):
    def __init__(
        self,
        delegate,
        loop: Any | None = ...,
        logger: Any | None = ...,
        name: str = ...,
    ) -> None: ...
    def submit(self, *args, **kwargs): ...
    def submit_with_loop(self, loop, fn, *args, **kwargs): ...
    def shutdown(self, wait: bool = ..., **_kwargs) -> None: ...



from .asyncio import AsyncioExecutor as AsyncioExecutor
from .bind import BoundCallable as BoundCallable
from .cancel_on_shutdown import CancelOnShutdownExecutor as CancelOnShutdownExecutor
from .flat_map import FlatMapExecutor as FlatMapExecutor
from .map import MapExecutor as MapExecutor
from .poll import PollExecutor as PollExecutor
from .retry import RetryExecutor as RetryExecutor
from .sync import SyncExecutor as SyncExecutor
from .throttle import ThrottleExecutor as ThrottleExecutor
from .timeout import TimeoutExecutor as TimeoutExecutor
from .wrapped import (
    CustomizableProcessPoolExecutor as CustomizableProcessPoolExecutor,
    CustomizableThreadPoolExecutor as CustomizableThreadPoolExecutor,
)

from .shared_types import ExecutorProtocol

class Executors:
    @classmethod
    def bind(cls, executor, fn): ...
    @classmethod
    def flat_bind(cls, executor, fn): ...
    @classmethod
    def thread_pool(cls, *args, **kwargs) -> ExecutorProtocol: ...
    @classmethod
    def process_pool(cls, *args, **kwargs): ...
    @classmethod
    def sync(cls, *args, **kwargs) -> SyncExecutor: ...
    @classmethod
    def with_retry(cls, executor, *args, **kwargs): ...
    @classmethod
    def with_map(cls, executor, *args, **kwargs): ...
    @classmethod
    def with_flat_map(cls, executor, *args, **kwargs): ...
    @classmethod
    def with_poll(cls, executor, *args, **kwargs): ...
    @classmethod
    def with_timeout(cls, executor, *args, **kwargs): ...
    @classmethod
    def with_throttle(cls, executor, *args, **kwargs): ...
    @classmethod
    def with_cancel_on_shutdown(cls, executor, *args, **kwargs): ...
    @classmethod
    def with_asyncio(cls, executor, *args, **kwargs): ...



import sys
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

from .wrap import CanCustomizeBind
from .metrics import metrics, track_future
from .helpers import ShutdownHelper


class CustomizableThreadPoolExecutor(CanCustomizeBind, ThreadPoolExecutor):
    def __init__(self, *args, **kwargs):
        self.__name = kwargs.pop("name", "default")
        self.__shutdown = ShutdownHelper()

        if (
            sys.version_info >= (3, 6)
            and len(args) < 2
            and "thread_name_prefix" not in kwargs
            and self.__name != "default"
        ):
            kwargs["thread_name_prefix"] = "ThreadPoolExecutor-%s" % self.__name

        super(CustomizableThreadPoolExecutor, self).__init__(*args, **kwargs)
        metrics.EXEC_TOTAL.labels(type="threadpool", executor=self.__name).inc()
        metrics.EXEC_INPROGRESS.labels(type="threadpool", executor=self.__name).inc()

    def shutdown(self, *args, **kwargs):
        if self.__shutdown():
            metrics.EXEC_INPROGRESS.labels(
                type="threadpool", executor=self.__name
            ).dec()
            super(CustomizableThreadPoolExecutor, self).shutdown(*args, **kwargs)

    def submit(self, *args, **kwargs):
        with self.__shutdown.ensure_alive():
            f = super(CustomizableThreadPoolExecutor, self).submit(*args, **kwargs)
            track_future(f, type="threadpool", executor=self.__name)
            return f


class CustomizableProcessPoolExecutor(CanCustomizeBind, ProcessPoolExecutor):
    def __init__(self, *args, **kwargs):
        kwargs.pop("name", None)
        super(CustomizableProcessPoolExecutor, self).__init__(*args, **kwargs)



from typing import TypeVar
import logging

from .shared_types import ExecutorProtocol

T = TypeVar("T")

class SyncExecutor(ExecutorProtocol):
    def __init__(
        self, logger: logging.Logger | None = ..., name: str = ...
    ) -> None: ...
    def __enter__(self) -> SyncExecutor: ...



# -*- coding: utf-8 -*-

from concurrent.futures import Future

from ..executors import Executors
from ..common import copy_exception
from ..metrics import track_future

EXECUTOR = Executors.sync(name="internal")


def f_return(x=None):
    """Return a future which provides the value `x`.

    Signature: :code:`A ⟶ Future<A>`

    Arguments:
        x
            A value to be returned

    Returns:
        :class:`~concurrent.futures.Future` of :obj:`x`
            A future immediately resolved with the value :obj:`x`.

    .. versionadded:: 1.19.0
    .. versionchanged:: 2.1.0
        value now defaults to :code:`None`
    """
    future = Future()
    track_future(future, type="return", executor="none")
    future.set_result(x)
    return future


def f_return_error(x, traceback=None):
    """Return a future which returns/raises the exception `x`.

    Arguments:

        x (Exception)
            An exception to be returned/raised.

        traceback (traceback)
            An optional traceback associated with the exception.

            This argument should only be provided on Python 2.x.
            It will be used in the value returned from
            :meth:`concurrent.futures.Future.exception_info`,
            which only exists on Python 2.

    Returns:
        :class:`~concurrent.futures.Future` of :obj:`x`
            A future immediately resolved with the exception :obj:`x`.

    .. versionadded:: 1.19.0
    """

    f = Future()
    track_future(f, type="return_error", executor="none")
    copy_exception(f, x, traceback)
    return f


def f_return_cancelled():
    """
    Returns:
        :class:`~concurrent.futures.Future`
            A future which is cancelled.

    .. versionadded:: 1.19.0
    """
    f = Future()
    track_future(f, type="return_cancelled", executor="none")
    f.cancel()
    f.set_running_or_notify_cancel()
    return f


class WeakCallback(object):
    # A wrapper for a single-call callback which breaks the reference
    # to the callback at time of call.
    #
    # The point is to avoid futures holding references to each other
    # unnecessarily.  Our own future base class already throws away
    # callbacks as they're called, but the standard concurrent.futures.Future
    # does not.
    def __init__(self, delegate):
        self.__delegate = delegate

    def __call__(self, *args, **kwargs):
        delegate = self.__delegate
        del self.__delegate
        return delegate(*args, **kwargs)


weak_callback = WeakCallback


def chain_cancel(f_outer, f_inner):
    # attempt to cancel f_inner when f_outer is cancelled
    f_outer.add_done_callback(
        weak_callback(lambda f: f_inner.cancel() if f.cancelled() else False)
    )


def wrap(f):
    return EXECUTOR.flat_bind(lambda: f)



from concurrent.futures import Future
from typing import TypeVar

T = TypeVar("T")

# TODO: is there any way we can express that the returned value
# will share most of the properties of T?
def f_proxy(f: Future[T], **kwargs) -> Future[T]: ...



from concurrent.futures import Future
from typing import Any, Callable, TypeVar, overload

A = TypeVar("A")
B = TypeVar("B")

@overload
def f_map(
    future: Future[A],
) -> Future[A]: ...
@overload
def f_map(
    future: Future[A],
    fn: Callable[[A], B] | None = ...,
    error_fn: Callable[[Exception], B] | None = ...,
) -> Future[B]: ...
@overload
def f_flat_map(
    future: Future[A],
) -> Future[A]: ...
@overload
def f_flat_map(
    future: Future[A],
    fn: Callable[[A], Future[B]] | None = ...,
    error_fn: Callable[[Exception], Future[B]] | None = ...,
) -> Future[B]: ...



from concurrent.futures import Executor

from .metrics import metrics
from .helpers import ShutdownHelper

try:
    import asyncio
except Exception:  # pylint: disable=broad-except
    pass


class AsyncioExecutor(Executor):
    """An executor which delegates to another executor while converting
    returned futures into instances of :class:`asyncio.Future`.

    Note that since this class produces :mod:`asyncio` rather than :mod:`concurrent.futures`
    future objects, AsyncioExecutor instances themselves cannot be used
    as a delegate executor of other executor instances within this library.

    .. versionadded:: 1.7.0
    """

    def __init__(self, delegate, loop=None, logger=None, name="default"):
        """
        Parameters:

            delegate (~concurrent.futures.Executor):
                executor to which callables will be submitted

            loop (~asyncio.AbstractEventLoop):
                asyncio event loop used to wrap futures; if omitted, the default
                event loop is used.

            logger (~logging.Logger):
                a logger used for messages from this executor

            name (str):
                a name for this executor

        .. versionchanged:: 2.7.0
            Introduced ``name``.
        """
        self._delegate = delegate
        self._loop = loop
        self._name = name
        self._shutdown = ShutdownHelper()
        metrics.EXEC_TOTAL.labels(type="asyncio", executor=self._name).inc()
        metrics.EXEC_INPROGRESS.labels(type="asyncio", executor=self._name).inc()

    def submit(self, *args, **kwargs):  # pylint: disable=arguments-differ
        return self.submit_with_loop(self._loop, *args, **kwargs)

    def submit_with_loop(self, loop, fn, *args, **kwargs):
        """Submit a callable with the specified event loop.

        Parameters:
            loop (~asyncio.AbstractEventLoop):
                asyncio event loop used to wrap futures

            fn (callable):
                callable to be submitted

        Returns:
            asyncio.Future:
                a future for the given callable
        """
        with self._shutdown.ensure_alive():
            if not loop:
                loop = asyncio.get_event_loop()
            future = self._delegate.submit(fn, *args, **kwargs)
            return asyncio.wrap_future(future, loop=loop)

    def shutdown(self, wait=True, **_kwargs):
        if self._shutdown():
            metrics.EXEC_INPROGRESS.labels(type="asyncio", executor=self._name).dec()
            self._delegate.shutdown(wait, **_kwargs)



from concurrent.futures import Future
from typing import TypeVar, overload

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")

@overload
def f_zip() -> Future[tuple]: ...
@overload
def f_zip(f1: Future[A]) -> Future[tuple[A]]: ...
@overload
def f_zip(f1: Future[A], f2: Future[B]) -> Future[tuple[A, B]]: ...
@overload
def f_zip(f1: Future[A], f2: Future[B], f3: Future[C]) -> Future[tuple[A, B, C]]: ...

# TODO: is there any way to make this work perfectly for arbitrary number of futures?
@overload
def f_zip(
    f1: Future, f2: Future, f3: Future, f4: Future, *fs: Future
) -> Future[tuple]: ...
